/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  cache.c -- Stuff for requesting files through an external cache (eg. squid)

  Current Problems include: 
      o No support for record structure.
      o STAT does nothing useful.
      o The wget command is compiled in.
  ***************************************/


#include <sys/signal.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "control.h"
#include "common.h"
#include "cache.h"


void cache_stat(const unsigned char *cmd,const unsigned char *arg);
void cache_abor(const unsigned char *cmd,const unsigned char *arg);
void cache_quit(const unsigned char *cmd,const unsigned char *arg);
void cache_mode(const unsigned char *cmd,const unsigned char *arg);
void cache_stru(const unsigned char *cmd,const unsigned char *arg);
void cache_type(const unsigned char *cmd,const unsigned char *arg);
void cache_rest(const unsigned char *cmd,const unsigned char *arg);
void cache_retrieve(const unsigned char *cmd,const unsigned char *arg);
void got_cwd(const unsigned char *msg);

static RETSIGTYPE cache_signal(int signo);

static unsigned char filename[MAX_LINE_LEN];
static int working=FALSE;
static int anon=FALSE;
static int type=0;
static pid_t wget=0;
static int aborting=0;
static int noop=0;

/* same as parse_cmd, but intercepts those commands we need to play with for
   doing RETR through an external program. Returns TRUE if it has dealt with
   the command, and FALSE if parse_cmd should deal as usual.*/
int cache_parsed_cmd(const unsigned char *cmd, const unsigned char *arg)
{
	if(!config_usecache()) return(FALSE);

	if(!strncasecmp(cmd,"USER",4)) {
		if(!strncasecmp(arg,"ftp",3) || !strncasecmp(arg,"anonymous",9))
			anon=TRUE;
		else anon=FALSE;
		return(FALSE);
	}

	if(!anon) return(FALSE);

	/*Commands that we need to change*/
	if(!strncasecmp(cmd,"MODE",4)) cache_mode(cmd,arg);
	else if(!strncasecmp(cmd,"STRU",4)) cache_stru(cmd,arg);
	else if(!strncasecmp(cmd,"TYPE",4)) cache_type(cmd,arg);
	else if(!strncasecmp(cmd,"REST",4)) cache_rest(cmd,arg);

	/*The retrieve command*/
	else if(!strncasecmp(cmd,"RETR",4)) cache_retrieve(cmd,arg);

	/*Commands we only change during our transfer*/
	else{
		if(!working) return(FALSE);

		if(!strncasecmp(cmd,"STAT",4)) cache_stat(cmd,arg);
		else if(!strncasecmp(cmd,"ABOR",4)) cache_abor(cmd,arg);
		else if(!strncasecmp(cmd,"QUIT",4)) cache_quit(cmd,arg);
	
		/*Otherwise we don't process it*/
		else return(FALSE);
	}

	return(TRUE);
}

/*Deal with server replies if we need to. Note that we have to deal
  with replies to our NOOPs even if we aren't working as wget may finish
  in between sending NOOP and getting the reply. */
int cache_parsed_reply(int code, const unsigned char *msg)
{
	if(!config_usecache()) return(FALSE);
	if(noop>0){
		if(code>0) noop--;
		return(TRUE);
	}
	if(!working) return(FALSE);

	switch(code){
	case 257:
		got_cwd(msg);
		break;
	default:
	}
	return(TRUE);
}		
	

void cache_stat(const unsigned char *cmd,const unsigned char *arg)
{
	send_message(213, "Retrieving file through cache");
}

void cache_abor(const unsigned char *cmd,const unsigned char *arg)
{
	aborting=1;
	kill(wget,SIGTERM);
	send_message(426, "transfer aborted");
	send_message(226, "Closing data connection");
}

void cache_quit(const unsigned char *cmd,const unsigned char *arg)
{
	kill(wget, SIGTERM);

	send_command(cmd, arg);
	return;
}

void cache_mode(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "S", 1)) send_message(200, "Command okay");
	else send_message(504,"Only stream mode implemented");
}

void cache_stru(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "F", 1)) send_message(200, "Command okay");
	else send_message(504,"Only file structure implemented");
}

void cache_type(const unsigned char *cmd,const unsigned char *arg)
{
	if(!strncasecmp(arg, "I", 1)) {
		type=1;
		send_command(cmd,arg);
	}
	else if(!strncasecmp(arg, "AN", 1) || !strncasecmp(arg,"A",1)){
		type=0;
		send_command(cmd,arg);
	}
	else send_message(504,"Only types I and AN implemented");
}

void cache_rest(const unsigned char *cmd,const unsigned char *arg)
{
	send_message(502,"Command not implemented.");
}

void cache_retrieve(const unsigned char *cmd,const unsigned char *arg)
{
	working=TRUE;
	strncpy(filename, arg, MAX_LINE_LEN-1);

	/*Get the working directory*/
	send_command("PWD","");
}

/* And once we have the cwd get the file*/
void got_cwd(const unsigned char *msg)
{
	unsigned char pwd[MAX_LINE_LEN], proxy[MAX_LINE_LEN], *tp;
	unsigned char buf[BUF_LEN];
	int fds[2];
	struct in_addr addr;

	/* FIXME We need to do some checking of the filename here
           before we pass it to wget*/

	/*Set up the wget command*/
	strncpy(pwd,strchr(msg, '"')+1,MAX_LINE_LEN-1);
	tp=strchr(pwd,'"');
	if(tp==NULL) exit(1);
	*(tp++)='/';
	*tp=0;

	addr.s_addr=info->server_control.address.addr;
	sprintf(buf,"ftp://%s/%s%s",
		inet_ntoa(addr),
		pwd, filename);
	if(type==0) strcat(buf,";type=a");
	debug2("ftp string = %s\n", buf);

	sprintf(proxy,"ftp_proxy=%s", config_ftpproxy());

	/*Exec wget*/
	signal(SIGCHLD, cache_signal);
	signal(SIGALRM, cache_signal);
	pipe(fds);
	wget=fork();

	switch(wget){
	case -1:
		exit(1);
	case 0:
		close(1);
		dup(fds[1]);
		close(fds[0]);
		close(fds[1]);
		execlp("wget","wget","-q","-e",proxy,"-O","-", buf, NULL);
		break;
	default:
		close(fds[1]);
		if(info->state & SERVERDATA) close(info->server_data.fd);
		info->server_data.fd=fds[0];
		send_message(150,"Starting transfer");
	}

	/*And set everything up as if the connection was normal*/
	if( !(info->state & CLIENTDATA) ){
		if(info->state & ACTIVE){
			info->client_data.fd=connect_to_socket(info->client_data.address);
			info->state |= CLIENTDATA;
		}
	}

	info->server_data.buflen=info->client_data.buflen=0;
	info->state |=SERVERDATA;

        fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
        fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);	

	/*Finally set up our timer for keeping server alive.*/
	alarm(30); /*FIXME make this configureable*/
}

static RETSIGTYPE cache_signal(int signo)
{
	int status;

	switch(signo){
	case SIGCHLD: /*wget finished - we need to disconnect*/
		working=FALSE;
		wget=0;
		waitpid(-1, &status, WNOHANG);
		if(WIFEXITED(status))
			write(info->client_control.fd,"226 Transfer complete\r\n",23);
		else if(!aborting) write(info->client_control.fd,"426 File transfer error\r\n",25);
		aborting=0;
		alarm(0);
		break;
	case SIGALRM: /*Send a NOOP to let the server know we are still here*/
		write(info->server_control.fd,"NOOP\r\n",6);
		noop++;
		alarm(30); /*FIXME make this configureable*/
		break;
	}
}
