/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  config.c: configuration file parsing
  
  ***************************************/


#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>

#include "common.h"

void usage(int exitval);

struct options {
	unsigned char listen_name[251];
	unsigned char config_file[251];
	int maxforks;
	u_int32_t listen_address; /* Kept in network order */
	u_int16_t port;   /* Kept in host order */
	unsigned char device[251];
	unsigned char user[251];
	unsigned char group[251];
	uid_t uid;
	gid_t gid;
	int same_address;
	int bounce_defend;
	unsigned char chroot[251];
	unsigned char logfile[251];
	char ftp_proxy[251];
	struct timeval timeout;
};

static struct options config;

int process_cmdline(int argc, char *argv[])
{
	int opt;

	memset(&config, 0, sizeof(config));
	strcpy(config.config_file,CONFIG_FILE);

	while((opt=getopt(argc,argv, "hf:")) !=-1) {
		switch(opt) {
		case 'f':
			strncpy(config.config_file, optarg, 250);
			break;
		case 'h':
			usage(0);
		case ':':
			fprintf(stderr, "Option -%c needs an argument\n", optopt);
			usage(-1);
		case '?':
			fprintf(stderr, "Unknown option -%c\n", optopt);
			usage(-1);
		}
	}
	if (optind!=argc) return(-1);
	return(0);
}

void usage(int exitval)
{	
	fprintf(stderr,"Usage frox [-f config_file] [-n]\n");

	exit(exitval);
}


int read_config(void)
{
	FILE *fp;
	unsigned char buf[BUF_LEN];
	unsigned char *p;

	
	if ( (fp=fopen(config.config_file,"r")) == NULL ) return(-1);

	do{
		/* 
		** Remove leading whitespace. Skip comments and empty lines.
		*/
		fgets(buf,BUF_LEN,fp);
		for(p=buf;*p==' ' || *p=='\t';p++);
		if(*p=='#') continue;
		memmove(buf, p, strlen(p) +1);
		if ((p = strchr(buf, '\n')) != NULL) *p = '\0';
		if(*buf=='\0') continue;

		p=strtok(buf, " \t=");

		if(!strncmp(p, "Listen", 20))
			strncpy(config.listen_name, strtok(NULL, " \t="), 250);

		else if(!strncmp(p, "Port", 20))
			config.port = strtol(strtok(NULL, " \t="), NULL, 10);
		
		else if(!strncmp(p, "User", 20))
			strncpy(config.user, strtok(NULL, " \t="), 250);

		else if(!strncmp(p, "Group", 20))
			strncpy(config.group, strtok(NULL, " \t="), 250);

                else if(!strncmp(p, "Chroot", 20))
                        strncpy(config.chroot, strtok(NULL, " \t="), 250);

                else if(!strncmp(p, "LogFile", 20))
                        strncpy(config.logfile, strtok(NULL, " \t="), 250);

                else if(!strncmp(p, "FtpProxy", 20))
			strncpy(config.ftp_proxy, strtok(NULL, " \t="), 250);

                else if(!strncmp(p, "BindToDevice", 20))
			strncpy(config.device, strtok(NULL, " \t="), 250);

                else if(!strncmp(p, "MaxForks", 20))
			config.maxforks=atoi(strtok(NULL, " \t="));

                else if(!strncmp(p, "Timeout", 20)){
			config.timeout.tv_sec=atoi(strtok(NULL, " \t="));
			config.timeout.tv_usec=0;
		}
                else if(!strncmp(p, "SameAddress", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				config.same_address=TRUE;
			else config.same_address=FALSE;
		}
                else if(!strncmp(p, "BounceDefend", 20))
		{
			if(!strncasecmp(strtok(NULL, " \t="), "yes", 4) ) 
				config.bounce_defend=TRUE;
			else config.bounce_defend=FALSE;
		}

	}while(!feof(fp));

	fclose(fp);

	if(*config.logfile!=0){
		int tmpfd;
		tmpfd=open(config.logfile, O_APPEND|O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
		if(tmpfd==-1 || dup2(tmpfd,2)==-1){
			fprintf(stderr, "Unable to open logfile %s\n",config.logfile);
			exit(-1);
		}
	}
		
	if(config.timeout.tv_sec==0){
		config.timeout.tv_sec=3600;
		config.timeout.tv_usec=0;
	}
	if(*config.listen_name == 0) 
		config.listen_address = htonl(INADDR_ANY);
	else if(*config.listen_name < '0' || *config.listen_name >'9' ) {
		struct hostent *hostinfo;

		hostinfo = gethostbyname(config.listen_name);
		if(hostinfo == NULL){
			fprintf(stderr, "Unable to resolve host %s\n", config.listen_name);
			return(-1);
		}
		config.listen_address = ((struct in_addr *) *hostinfo->h_addr_list)->s_addr;
	}
	else config.listen_address = inet_addr(config.listen_name);

	/* We need to find out the uid / gid now. We must chroot
	** before changing priveleges, but if we wait until after
	** chroot we cannot parse /etc/passwd. 
	*/

	if(*config.group!=0){
		struct group* grp;
		if( (grp=getgrnam(config.group)) ==NULL){
			fprintf(stderr, "Unable to find group %s\n",config.group);
			return(-1);
		}
		config.gid=grp->gr_gid;
	}

	if(*config.user!=0){
		struct passwd* pw;
		if( (pw=getpwnam(config.user)) == NULL){
			fprintf(stderr, "Unable to find user %s\n",config.user);
			return(-1);
		}
		config.uid=pw->pw_uid;
	}
	return(0);
}	

u_int32_t config_getaddress()
{
	return(config.listen_address);
}

u_int16_t config_getport()
{
	return(htons(config.port));
}

int config_maxforks()
{
	return(config.maxforks);
}

struct timeval config_timeout(void)
{
	return(config.timeout);
}

int config_usesameaddress()
{
	return(config.same_address);
}

int config_bouncedefend()
{
	return(config.bounce_defend);
}

int config_bindtodevice(int fd)
{
	if(*config.device==0) return(0);
	if(setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, 
		      (void *)config.device, (socklen_t) strlen(config.device)+1) != 0){
		perror("Binding to device");
		return(-1);
	}
	write_log(INFO, "Bound to device %s",config.device);
	return(0);
}
	

int config_chroot()
{
	if(*config.chroot==0) return(0);
	if( chroot(config.chroot)!=0 ) return(-1);
	if( chdir("/") !=0) return(-1);
	write_log(INFO, "Chrooted to %s", config.chroot);
	return(0);
}

int config_droppriv()
{
	if(config.gid!=0){
		setgid(config.gid);
		setgid(config.gid);
	}
	if(config.uid!=0){
		setuid(config.uid);
		setuid(config.uid);
		write_log(INFO, "Dropped priveledges");
	}
	return(0);
}

int config_usecache()
{
	if(*config.ftp_proxy==0) return(FALSE);
	else return(TRUE);
}

const char* config_ftpproxy()
{
	return(config.ftp_proxy);
}
